/*****************************************************************************
 embedded Filesystem
 *****************************************************************************/

/*
eigentlich liegt hier noch alles im Argen, denn die Herangehensweise bei
Chan's FF pat nicht so recht. Eigentlich mte man in "InitPartitionRecord"
jeder gefundenen Partiton ein Filesystem zuordnen, also bei FATxx eben Chan's
und bei z.B. 83h ein ext2 FS oder so. Aber Chan's FF will alles lesen, vom MBR
angefangen. Dazu kommt noch ne Marotte mit Laufwerksnummern, also
A:\MyDir\blabla.exe   wird hier zu   0:/MyDir/blabla.exe

*/

#include "LPC24xx.h"
#include "StdTypes.h"
#include "startup.h"
#include "conv.h"
#include "serial.h"
#include "ff.h"
#include "diskio.h"
#include "sdio.h"
#include "efsys.h"


FATFS    Fatfs;        // File system object

// 4 vorgehaltene Partitonen
TPARTREC  PartRec [4];

byte buf[512];

// MBR lesen und PartRec aufsetzen
bool InitPartitionRecord (void)
{
  bool  b;
  int   i;

  i = 4;
  while(i) { --i; PartRec[i].Magic = 0; }  // alle austragen

  b = SD_ReadBlock (0, (dword*) &buf);
  if (!b) return false;
  if ((buf[510]!=0x55)||(buf[511]!=0xAA)) return false;

  // Partitionen aus MBR lesen
  i = 0;
  while(i<4)
  { PartRec[i].Magic       = buf[(i*16)+446+4];
    PartRec[i].ReadOnly    = SD_readonly;
    PartRec[i].DriveLetter = 'A'+i;
    PartRec[i].Changed     = false;
    PartRec[i].LBA_Start   = GetDWord(buf,(i*16)+446+8);
    PartRec[i].LBA_NumSec  = GetDWord(buf,(i*16)+446+12);

    ++i;
  }

  // bei gefundenen Partitionen Bootsektor auslesen
  i = 0;
  while(i<4)
  { if ((PartRec[i].Magic==6) || (PartRec[i].Magic==13))
    { b = SD_ReadBlock (0, (dword*) &buf);
      if (!b) return false;
      if ((buf[510]!=0x55)||(buf[511]!=0xAA)) return false;
    }
    ++i;
  }

  f_mount(0, &Fatfs);
  return true;
}

/*****************************************************************************
 Anschlu an Chan's FF
 *****************************************************************************/



/* Initialize a Drive  */
DSTATUS disk_initialize (BYTE pdrv)     /* Physical drive nmuber (0..) */
{ if (pdrv!=0) return STA_NODISK;

  if ((SDCardTyp == SD1_1) ||
      (SDCardTyp == SD2_0) ||
      (SDCardTyp == SDHC))
  { if (SD_readonly) return STA_PROTECT;
    return 0;
  }
  return STA_NODISK;
}



/*-----------------------------------------------------------------------*/
/* Get Disk Status                                                       */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status (BYTE pdrv)		/* Physical drive nmuber (0..) */
{ if (pdrv!=0) return STA_NODISK;
  if ((SDCardTyp == SD1_1) ||
      (SDCardTyp == SD2_0) ||
      (SDCardTyp == SDHC))
  { if (SD_readonly) return STA_PROTECT;
    return 0;
  }
  return STA_NODISK;
}


/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

DRESULT disk_read (
	BYTE pdrv,		/* Physical drive number (0..) */
	BYTE *buff,		/* Data buffer to store read data */
	DWORD sector,	/* Sector address (LBA) */
	BYTE count)		/* Number of sectors to read (1..128) */
{ bool b;

  if (pdrv!=0) return RES_PARERR;
  while (count)
  { b = SD_ReadBlock (sector, (dword*) buff);
    if (!b) return RES_ERROR;
    buff += 512;
    ++sector;
    --count;
  }
  return RES_OK;
}



/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

#if _USE_WRITE
DRESULT disk_write (
	BYTE pdrv,		/* Physical drive nmuber (0..) */
	const BYTE *buff,	/* Data to be written */
	DWORD sector,		/* Sector address (LBA) */
	BYTE count)		/* Number of sectors to write (1..128) */

{ bool b;

  if (pdrv!=0) return RES_PARERR;
  while (count)
  { b = SD_WriteBlock (sector, (dword*) buff);
    if (!b) return RES_ERROR;
    buff += 512;
    ++sector;
    --count;
  }
  return RES_OK;
}
#endif


/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

#if _USE_IOCTL
DRESULT disk_ioctl (
	BYTE pdrv,		/* Physical drive number (0..) */
	BYTE cmd,		/* Control code */
	void *buff)		/* Buffer to send/receive control data */

{ WORD* w;
  DWORD* d;

  if (pdrv!=0) return RES_PARERR;

  if (cmd==GET_SECTOR_SIZE)
  { w = buff;
    *w = 512;
    return RES_OK;
  }
  if (cmd==GET_SECTOR_COUNT)
  { d = buff;
    *d = SD_CardSize/512;
    return RES_OK;
  }
  if (cmd==GET_BLOCK_SIZE)
  { d = buff;
    *d = 1;
    return RES_OK;
  }
  return RES_PARERR;
}
#endif


/******************************************************************************
 Hilfsroutinen
 ******************************************************************************/
word  GetWord  (byte* buf, int aIndex)    // lesen eines 16 Bit Wortes auf krummen Adressen
{ word w;
  w = buf[aIndex+1];
  w = (w<<8) | buf[aIndex];
  return w;
}

dword GetDWord (byte* buf, int aIndex)    // lesen eines 32 Bit Wortes auf krummen Adressen
{ dword d;
  d = buf[aIndex+3];
  d = (d<<8) | buf[aIndex+2];
  d = (d<<8) | buf[aIndex+1];
  d = (d<<8) | buf[aIndex];
  return d;
}

#define jahr   2013
#define monat     4
#define tag       8

dword get_fattime (void)
{ dword d;
  dword sek, minu, stun;

  sek  = T0TC / 1000;
  minu = sek / 60;
  stun = minu / 60;
  stun = stun % 24;
  minu = minu % 60;
  sek  = sek % 60;
  d = ((jahr - 1980) << 25) |
      (monat << 21) |
      (tag   << 16) |
      (stun  << 11) |
      (minu  <<  5) |
      (sek   >>  1);
  return d;
}



char* TranslateErrorcode(FRESULT code)
{ switch (code)
  { case FR_OK:                  return (char*) "ok.";
    case FR_DISK_ERR:            return (char*) "allgemeiner Disk Fehler";
    case FR_INT_ERR:             return (char*) "interner Fehler";
    case FR_NOT_READY:           return (char*) "Medium nicht bereit";
    case FR_NO_FILE:             return (char*) "keine solche Datei";
    case FR_NO_PATH:             return (char*) "Pfad nicht gefunden";
    case FR_INVALID_NAME:        return (char*) "ungltiger Name";
    case FR_DENIED:              return (char*) "Zugriff verweigert";
    case FR_EXIST:               return (char*) "Datei/Pfad existiert bereits";
    case FR_WRITE_PROTECTED:     return (char*) "Medium ist R/O";
    case FR_INVALID_DRIVE:       return (char*) "ungltiges Laufwerk";
    case FR_NOT_ENABLED:         return (char*) "Laufwerk ist nicht gemounted";
    case FR_NO_FILESYSTEM:       return (char*) "kein FAT Filesystem!";
    case FR_TIMEOUT:             return (char*) "Timeout";
    case FR_LOCKED:              return (char*) "Datei ist gesperrt";
    case FR_NOT_ENOUGH_CORE:     return (char*) "zu wenig RAM!";
    case FR_TOO_MANY_OPEN_FILES: return (char*) "zu viele offene Dateien";
    case FR_INVALID_PARAMETER:   return (char*) "ungltiger Parameter";
  }
  return "allg. Fehler";
}
